@@ -37,6 +37,8 @@ module External : sig
   val cwd : unit -> t
 
   val as_local : t -> string
+
+  val of_filename_relative_to_initial_cwd : string -> t
 end = struct
   module Table = String.Table
 
@@ -63,7 +65,7 @@ end = struct
       User_error.raise ~loc [ Pp.textf "path %s is not absolute" t ];
     t
 
-  let to_dyn t = Dyn.String t
+  let to_dyn t = Dyn.variant "External" [ Dyn.string t ]
 
   (* let rec cd_dot_dot t = match Unix.readlink t with | exception _ ->
      Filename.dirname t | t -> cd_dot_dot t
@@ -110,6 +112,10 @@ end = struct
     let base, _ = split_extension t in
     base ^ ext
 
+  let map_extension t ~f =
+    let base, ext = split_extension t in
+    base ^ f ext
+
   let cwd () = Sys.getcwd ()
 
   let initial_cwd = Fpath.initial_cwd
@@ -118,6 +124,9 @@ end = struct
     let s = t in
     "." ^ s
 
+  let of_filename_relative_to_initial_cwd fn =
+    if Filename.is_relative fn then relative initial_cwd fn else of_string fn
+
   include (
     Comparator.Operators (struct
       type nonrec t = t
@@ -218,11 +227,9 @@ end = struct
         | [] -> Result.Ok t
         | "." :: rest -> loop t rest
         | ".." :: rest -> (
-          if is_root t then Result.Error ()
-          else
-            match parent t with
-            | None -> Error ()
-            | Some parent -> loop parent rest)
+          match parent t with
+          | None -> Result.Error `Outside_the_workspace
+          | Some parent -> loop parent rest)
         | fn :: rest ->
           if is_root t then loop fn rest else loop (t ^ "/" ^ fn) rest
       in
@@ -231,7 +238,7 @@ end = struct
     let relative ?error_loc t components =
       match relative_result t components with
       | Result.Ok t -> t
-      | Error () ->
+      | Error `Outside_the_workspace ->
         User_error.raise ?loc:error_loc
           [ Pp.textf "path outside the workspace: %s from %s"
               (String.concat ~sep:"/" components)
@@ -245,7 +252,7 @@ end = struct
         [ ("t", to_dyn t); ("path", String path) ];
     match L.relative_result t (explode_path path) with
     | Result.Ok t -> t
-    | Error () ->
+    | Error `Outside_the_workspace ->
       User_error.raise ?loc:error_loc
         [ Pp.textf "path outside the workspace: %s from %s" path t ]
 
@@ -343,6 +350,10 @@ end = struct
     let base, _ = split_extension t in
     base ^ ext
 
+  let map_extension t ~f =
+    let base, ext = split_extension t in
+    base ^ f ext
+
   module Prefix = struct
     type _ t =
       { len : int
@@ -421,6 +432,9 @@ module Local : sig
 
   module L : sig
     val relative : ?error_loc:Loc0.t -> t -> string list -> t
+
+    val relative_result :
+      t -> string list -> (t, [ `Outside_the_workspace ]) Result.t
   end
 
   val split_first_component : t -> (string * t) option
@@ -483,7 +497,11 @@ module Relative_to_source_root = struct
     ignore (Fpath.mkdir_p ?perms (Local.to_string path) : Fpath.mkdir_p_result)
 end
 
-module Source0 = Local
+module Source0 = struct
+  include Local
+
+  let to_dyn s = Dyn.variant "In_source_tree" [ to_dyn s ]
+end
 
 let abs_root, set_root =
   let root_dir = ref None in
@@ -505,7 +523,7 @@ let abs_root, set_root =
   in
   (abs_root, set_root)
 
-module Kind = struct
+module Outside_build_dir = struct
   type t =
     | External of External.t
     | In_source_dir of Local.t
@@ -521,7 +539,9 @@ module Kind = struct
     | In_source_dir t -> Local.to_string t
     | External t -> External.to_string t
 
-  let to_dyn t = Dyn.String (to_string t)
+  let to_dyn = function
+    | In_source_dir t -> Source0.to_dyn t
+    | External t -> External.to_dyn t
 
   let of_string s =
     if Filename.is_relative s then In_source_dir (Local.of_string s)
@@ -531,10 +551,51 @@ module Kind = struct
     | In_source_dir t -> Relative_to_source_root.mkdir_p ?perms t
     | External t -> External.mkdir_p ?perms t
 
+  let relative t s =
+    match t with
+    | In_source_dir t -> In_source_dir (Local.relative t s)
+    | External t -> External (External.relative t s)
+
+  let extend_basename t ~suffix =
+    match t with
+    | In_source_dir t -> In_source_dir (Local.extend_basename t ~suffix)
+    | External t -> External (External.extend_basename t ~suffix)
+
   let append_local x y =
     match x with
     | In_source_dir x -> In_source_dir (Local.append x y)
     | External x -> External (External.relative x (Local.to_string y))
+
+  let to_string_maybe_quoted t = String.maybe_quoted (to_string t)
+
+  let equal (x : t) (y : t) =
+    match (x, y) with
+    | External x, External y -> External.equal x y
+    | External _, In_source_dir _ -> false
+    | In_source_dir x, In_source_dir y -> Local.equal x y
+    | In_source_dir _, External _ -> false
+
+  let hash = Poly.hash
+
+  let parent = function
+    | In_source_dir t -> (
+      match Local.parent t with
+      | None -> None
+      | Some s -> Some (In_source_dir s))
+    | External t -> (
+      match External.parent t with
+      | None -> None
+      | Some s -> Some (External s))
+
+  module Table = Hashtbl.Make (struct
+    type nonrec t = t
+
+    let hash = Poly.hash
+
+    let equal = Poly.equal
+
+    let to_dyn = to_dyn
+  end)
 end
 
 module Permissions = struct
@@ -619,11 +680,11 @@ module Build = struct
       Code_error.raise "Path.Build.drop_build_context_maybe_sandboxed_exn"
         [ ("t", to_dyn t) ]
 
-  let build_dir = Fdecl.create Kind.to_dyn
+  let build_dir = Fdecl.create Outside_build_dir.to_dyn
 
   let build_dir_prefix = Fdecl.create Dyn.opaque
 
-  let set_build_dir (new_build_dir : Kind.t) =
+  let set_build_dir (new_build_dir : Outside_build_dir.t) =
     let new_build_dir_prefix =
       (match new_build_dir with
       | External _ -> ()
@@ -650,6 +711,8 @@ module Build = struct
       if Local.is_root p then External.to_string b
       else Filename.concat (External.to_string b) (Local.to_string p)
 
+  let to_string_maybe_quoted p = String.maybe_quoted (to_string p)
+
   let of_local t = t
 
   let chmod t ~mode = Unix.chmod (to_string t) mode
@@ -658,7 +721,7 @@ module Build = struct
 
   let unlink_no_err t = Fpath.unlink_no_err (to_string t)
 
-  module Kind = Kind
+  let to_dyn s = Dyn.variant "In_build_dir" [ to_dyn s ]
 end
 
 module T : sig
@@ -669,6 +732,9 @@ module T : sig
 
   val to_dyn : t -> Dyn.t
 
+  (** [External _] < [In_source_tree _] < [In_build_dir _]
+
+      Path of the same kind are compared using the standard lexical order *)
   val compare : t -> t -> Ordering.t
 
   val equal : t -> t -> bool
@@ -692,13 +758,13 @@ end = struct
     | External _, _ -> Lt
     | _, External _ -> Gt
     | In_source_tree x, In_source_tree y -> Local.compare x y
-    | In_source_tree _, _ -> Lt
-    | _, In_source_tree _ -> Gt
+    | In_source_tree _, In_build_dir _ -> Lt
+    | In_build_dir _, In_source_tree _ -> Gt
     | In_build_dir x, In_build_dir y -> Local.compare x y
 
   let equal (x : t) (y : t) = x = y
 
-  let hash = Hashtbl.hash
+  let hash = Poly.hash
 
   let in_build_dir s = In_build_dir s
 
@@ -706,17 +772,15 @@ end = struct
 
   let external_ e = External e
 
-  let to_dyn t =
-    let open Dyn in
-    match t with
-    | In_build_dir s -> Variant ("In_build_dir", [ Local.to_dyn s ])
-    | In_source_tree s -> Variant ("In_source_tree", [ Local.to_dyn s ])
-    | External s -> Variant ("External", [ External.to_dyn s ])
+  let to_dyn = function
+    | In_build_dir s -> Build.to_dyn s
+    | In_source_tree s -> Source0.to_dyn s
+    | External s -> External.to_dyn s
 end
 
 include T
 
-let hash (t : t) = Hashtbl.hash t
+let hash (t : t) = Poly.hash t
 
 let build_dir = in_build_dir Local.root
 
@@ -724,10 +788,11 @@ let is_root = function
   | In_source_tree s -> Local.is_root s
   | In_build_dir _ | External _ -> false
 
-let kind = function
-  | In_build_dir p -> Kind.append_local (Fdecl.get Build.build_dir) p
-  | In_source_tree s -> Kind.In_source_dir s
-  | External s -> Kind.External s
+let local_or_external : t -> Outside_build_dir.t = function
+  | In_build_dir p ->
+    Outside_build_dir.append_local (Fdecl.get Build.build_dir) p
+  | In_source_tree s -> In_source_dir s
+  | External s -> External s
 
 let is_managed = function
   | In_build_dir _ | In_source_tree _ -> true
@@ -756,7 +821,16 @@ let relative ?error_loc t fn =
   | _ when not (Filename.is_relative fn) -> external_ (External.of_string fn)
   | _ -> (
     match t with
-    | In_source_tree p -> make_local_path (Local.relative p fn ?error_loc)
+    | In_source_tree p -> (
+      let fn' = explode_path fn in
+      match Local.L.relative_result p fn' with
+      | Ok l -> make_local_path l
+      | Error `Outside_the_workspace ->
+        external_
+          (External.relative
+             (External.of_string
+                (Outside_build_dir.to_absolute_filename (local_or_external t)))
+             fn))
     | In_build_dir p -> in_build_dir (Local.relative p fn ?error_loc)
     | External s -> external_ (External.relative s fn))
 
@@ -770,19 +844,11 @@ let parse_string_exn ~loc s =
 
 let of_string s = parse_string_exn ~loc:Loc0.none s
 
-let to_dyn =
-  let open Dyn in
-  function
-  | In_build_dir s -> variant "In_build_dir" [ Local.to_dyn s ]
-  | In_source_tree s -> variant "In_source_tree" [ Local.to_dyn s ]
-  | External s -> variant "External" [ External.to_dyn s ]
-
 let of_filename_relative_to_initial_cwd fn =
-  external_
-    (if Filename.is_relative fn then External.relative External.initial_cwd fn
-    else External.of_string fn)
+  external_ (External.of_filename_relative_to_initial_cwd fn)
 
-let to_absolute_filename t = Kind.to_absolute_filename (kind t)
+let to_absolute_filename t =
+  Outside_build_dir.to_absolute_filename (local_or_external t)
 
 let external_of_local x ~root =
   External.to_string (External.relative root (Local.to_string x))
@@ -884,6 +950,22 @@ let as_in_source_tree_exn t =
       "[as_in_source_tree_exn] called on something not in source tree"
       [ ("t", to_dyn t) ]
 
+let as_outside_build_dir_exn : t -> Outside_build_dir.t = function
+  | In_source_tree s -> In_source_dir s
+  | External s -> External s
+  | In_build_dir path ->
+    Code_error.raise "as_outside_build_dir_exn" [ ("path", Build.to_dyn path) ]
+
+let destruct_build_dir :
+    t -> [ `Inside of Build.t | `Outside of Outside_build_dir.t ] = function
+  | In_source_tree p -> `Outside (In_source_dir p)
+  | External s -> `Outside (External s)
+  | In_build_dir s -> `Inside s
+
+let outside_build_dir : Outside_build_dir.t -> t = function
+  | In_source_dir d -> In_source_tree d
+  | External s -> External s
+
 let as_in_build_dir = function
   | In_build_dir b -> Some b
   | In_source_tree _ | External _ -> None
@@ -896,13 +978,15 @@ let as_in_build_dir_exn t =
       [ ("t", to_dyn t) ]
   | In_build_dir p -> p
 
+let as_external = function
+  | External s -> Some s
+  | In_build_dir _ | In_source_tree _ -> None
+
 let extract_build_context = function
   | In_source_tree _ | External _ -> None
   | In_build_dir p when Local.is_root p -> None
   | In_build_dir t -> Build.extract_build_context t
 
-let extract_build_dir_first_component = extract_build_context
-
 let extract_build_context_exn t =
   match extract_build_context t with
   | Some t -> t
@@ -968,14 +1052,14 @@ let drop_optional_build_context_src_exn t =
   | In_source_tree p -> p
 
 let split_first_component t =
-  match kind t with
+  match local_or_external t with
   | In_source_dir t ->
     Option.map (Local.split_first_component t) ~f:(fun (before, after) ->
         (before, after |> in_source_tree))
   | _ -> None
 
 let explode t =
-  match kind t with
+  match local_or_external t with
   | In_source_dir p when Local.is_root p -> Some []
   | In_source_dir s -> Some (String.split (Local.to_string s) ~on:'/')
   | External _ -> None
@@ -985,6 +1069,17 @@ let explode_exn t =
   | Some s -> s
   | None -> Code_error.raise "Path.explode_exn" [ ("path", to_dyn t) ]
 
+let relative_to_source_in_build_or_external ?error_loc ~dir s =
+  match Build.extract_build_context dir with
+  | None -> relative ?error_loc (In_build_dir dir) s
+  | Some (bctxt, source) -> (
+    let path = relative ?error_loc (In_source_tree source) s in
+    match path with
+    | In_source_tree s ->
+      In_build_dir
+        (Build.relative (Build.of_string bctxt) (Source0.to_string s))
+    | In_build_dir _ | External _ -> path)
+
 let exists t = try Sys.file_exists (to_string t) with Sys_error _ -> false
 
 let readdir_unsorted t = Dune_filesystem_stubs.read_directory (to_string t)
@@ -1014,7 +1109,7 @@ let build_dir_exists () = is_directory build_dir
 
 let ensure_build_dir_exists () =
   let perms = 0o777 in
-  match kind build_dir with
+  match local_or_external build_dir with
   | In_source_dir p -> Relative_to_source_root.mkdir_p p ~perms
   | External p -> (
     let p = External.to_string p in
@@ -1055,7 +1150,8 @@ let mkdir_p ?perms = function
   | External s -> External.mkdir_p s ?perms
   | In_source_tree s -> Relative_to_source_root.mkdir_p s ?perms
   | In_build_dir k ->
-    Kind.mkdir_p ?perms (Kind.append_local (Fdecl.get Build.build_dir) k)
+    Outside_build_dir.mkdir_p ?perms
+      (Outside_build_dir.append_local (Fdecl.get Build.build_dir) k)
 
 let touch ?(create = true) p =
   let p =
@@ -1063,7 +1159,8 @@ let touch ?(create = true) p =
     | External s -> External.to_string s
     | In_source_tree s -> Local_gen.to_string s
     | In_build_dir k ->
-      Kind.to_string (Kind.append_local (Fdecl.get Build.build_dir) k)
+      Outside_build_dir.to_string
+        (Outside_build_dir.append_local (Fdecl.get Build.build_dir) k)
   in
   let create =
     if create then fun () -> Unix.close (Unix.openfile p [ Unix.O_CREAT ] 0o777)
@@ -1109,6 +1206,10 @@ let set_extension t ~ext =
   | In_build_dir t -> in_build_dir (Local.set_extension t ~ext)
   | In_source_tree t -> in_source_tree (Local.set_extension t ~ext)
 
+let map_extension t ~f =
+  let base, ext = split_extension t in
+  extend_basename ~suffix:(f ext) base
+
 module O = Comparable.Make (T)
 module Map = O.Map
 
@@ -1163,6 +1264,9 @@ let set_of_source_paths set =
 let set_of_build_paths_list =
   List.fold_left ~init:Set.empty ~f:(fun acc e -> Set.add acc (build e))
 
+let set_of_external_paths set =
+  External.Set.to_list set |> Set.of_list_map ~f:external_
+
 let rename old_path new_path =
   Sys.rename (to_string old_path) (to_string new_path)
 
@@ -1173,7 +1277,11 @@ let follow_symlink path =
 
 module Expert = struct
   let drop_absolute_prefix ~prefix p =
-    match String.drop_prefix ~prefix:(Kind.to_absolute_filename prefix) p with
+    match
+      String.drop_prefix
+        ~prefix:(Outside_build_dir.to_absolute_filename prefix)
+        p
+    with
     | None -> None
     | Some "" -> Some Local.root
     | Some p ->
@@ -1183,13 +1291,13 @@ module Expert = struct
     let p = External.to_string ext in
     match Fdecl.get Build.build_dir with
     | External s -> (
-      match drop_absolute_prefix ~prefix:(Kind.External s) p with
+      match drop_absolute_prefix ~prefix:(External s) p with
       | Some s -> Some (in_build_dir s)
       | None ->
-        drop_absolute_prefix ~prefix:(Kind.In_source_dir Local.root) p
+        drop_absolute_prefix ~prefix:(In_source_dir Local.root) p
         |> Option.map ~f:in_source_tree)
     | In_source_dir _ ->
-      drop_absolute_prefix ~prefix:(Kind.In_source_dir Local.root) p
+      drop_absolute_prefix ~prefix:(In_source_dir Local.root) p
       |> Option.map ~f:make_local_path
 
   let try_localize_external t =
